# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Tuple, Type

import torch
from torch import Tensor, nn

from .common import MLPBlock


class TwoWayTransformer(nn.Module):
    def __init__(
        self,
        depth: int,
        embedding_dim: int,
        num_heads: int,
        mlp_dim: int,
        activation: Type[nn.Module] = nn.ReLU,
        attention_downsample_rate: int = 2,
    ) -> None:
        """
        A transformer decoder that attends to an input image using
        queries whose positional embedding is supplied.

        Args:
          depth (int): number of layers in the transformer
          embedding_dim (int): the channel dimension for the input embeddings
          num_heads (int): the number of heads for multihead attention. Must
            divide embedding_dim
          mlp_dim (int): the channel dimension internal to the MLP block
          activation (nn.Module): the activation to use in the MLP block
        """
        super().__init__()
        self.depth = depth
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.mlp_dim = mlp_dim
        self.layers = nn.ModuleList()

        for i in range(depth):
            self.layers.append(
                TwoWayAttentionBlock(
                    embedding_dim=embedding_dim,
                    num_heads=num_heads,
                    mlp_dim=mlp_dim,
                    activation=activation,
                    attention_downsample_rate=attention_downsample_rate,
                    skip_first_layer_pe=(i == 0),
                )
            )

        self.final_attn_token_to_image = Attention(
            embedding_dim, num_heads, downsample_rate=attention_downsample_rate
        )
        self.norm_final_attn = nn.LayerNorm(embedding_dim)

    def forward(
        self,
        image_embedding: Tensor,
        image_pe: Tensor,
        point_embedding: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Args:
          image_embedding (torch.Tensor): image to attend to. Should be shape
            B x embedding_dim x h x w for any h and w.
          image_pe (torch.Tensor): the positional encoding to add to the image. Must
            have the same shape as image_embedding.
          point_embedding (torch.Tensor): the embedding to add to the query points.
            Must have shape B x N_points x embedding_dim for any N_points.

        Returns:
          torch.Tensor: the processed point_embedding
          torch.Tensor: the processed image_embedding
        """
        # BxCxHxW -> BxHWxC == B x N_image_tokens x C
        bs, c, h, w = image_embedding.shape
        image_embedding = image_embedding.flatten(2).permute(0, 2, 1)
        image_pe = image_pe.flatten(2).permute(0, 2, 1)

        # Prepare queries
        queries = point_embedding
        keys = image_embedding

        # Apply transformer blocks and final layernorm
        for layer in self.layers:
            queries, keys = layer(
                queries=queries,
                keys=keys,
                query_pe=point_embedding,
                key_pe=image_pe,
            )

        # Apply the final attenion layer from the points to the image
        q = queries + point_embedding
        k = keys + image_pe
        attn_out = self.final_attn_token_to_image(q=q, k=k, v=keys)
        queries = queries + attn_out
        queries = self.norm_final_attn(queries)

        return queries, keys


class TwoWayAttentionBlock(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        mlp_dim: int = 2048,
        activation: Type[nn.Module] = nn.ReLU,
        attention_downsample_rate: int = 2,
        skip_first_layer_pe: bool = False,
    ) -> None:
        """
        A transformer block with four layers: (1) self-attention of sparse
        inputs, (2) cross attention of sparse inputs to dense inputs, (3) mlp
        block on sparse inputs, and (4) cross attention of dense inputs to sparse
        inputs.

        Arguments:
          embedding_dim (int): the channel dimension of the embeddings
          num_heads (int): the number of heads in the attention layers
          mlp_dim (int): the hidden dimension of the mlp block
          activation (nn.Module): the activation of the mlp block
          skip_first_layer_pe (bool): skip the PE on the first layer
        """
        super().__init__()
        self.self_attn = Attention(embedding_dim, num_heads)
        self.norm1 = nn.LayerNorm(embedding_dim)

        self.cross_attn_token_to_image = Attention(
            embedding_dim, num_heads, downsample_rate=attention_downsample_rate
        )
        self.norm2 = nn.LayerNorm(embedding_dim)

        self.mlp = MLPBlock(embedding_dim, mlp_dim, activation)
        self.norm3 = nn.LayerNorm(embedding_dim)

        self.norm4 = nn.LayerNorm(embedding_dim)
        self.cross_attn_image_to_token = Attention(
            embedding_dim, num_heads, downsample_rate=attention_downsample_rate
        )

        self.skip_first_layer_pe = skip_first_layer_pe

    def forward(
        self, queries: Tensor, keys: Tensor, query_pe: Tensor, key_pe: Tensor
    ) -> Tuple[Tensor, Tensor]:
        # Self attention block
        if self.skip_first_layer_pe:
            queries = self.self_attn(q=queries, k=queries, v=queries)
        else:
            q = queries + query_pe
            attn_out = self.self_attn(q=q, k=q, v=queries)
            queries = queries + attn_out
        queries = self.norm1(queries)

        # Cross attention block, tokens attending to image embedding
        q = queries + query_pe
        k = keys + key_pe
        attn_out = self.cross_attn_token_to_image(q=q, k=k, v=keys)
        queries = queries + attn_out
        queries = self.norm2(queries)

        # MLP block
        mlp_out = self.mlp(queries)
        queries = queries + mlp_out
        queries = self.norm3(queries)

        # Cross attention block, image embedding attending to tokens
        q = queries + query_pe
        k = keys + key_pe
        attn_out = self.cross_attn_image_to_token(q=k, k=q, v=queries)
        keys = keys + attn_out
        keys = self.norm4(keys)

        return queries, keys


class Attention(nn.Module):
    """
    An attention layer that allows for downscaling the size of the embedding
    after projection to queries, keys, and values.
    """

    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        downsample_rate: int = 1,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.internal_dim = embedding_dim // downsample_rate
        self.num_heads = num_heads
        assert (
            self.internal_dim % num_heads == 0
        ), "num_heads must divide embedding_dim."

        self.q_proj = nn.Linear(embedding_dim, self.internal_dim)
        self.k_proj = nn.Linear(embedding_dim, self.internal_dim)
        self.v_proj = nn.Linear(embedding_dim, self.internal_dim)
        self.out_proj = nn.Linear(self.internal_dim, embedding_dim)

    def _separate_heads(self, x: Tensor, num_heads: int) -> Tensor:
        b, n, c = x.shape
        x = x.reshape(b, n, num_heads, c // num_heads)
        return x.transpose(1, 2)  # B x N_heads x N_tokens x C_per_head

    def _recombine_heads(self, x: Tensor) -> Tensor:
        b, n_heads, n_tokens, c_per_head = x.shape
        x = x.transpose(1, 2)
        return x.reshape(b, n_tokens, n_heads * c_per_head)  # B x N_tokens x C

    def forward(self, q: Tensor, k: Tensor, v: Tensor) -> Tensor:
        # Input projections
        q = self.q_proj(q)
        k = self.k_proj(k)
        v = self.v_proj(v)

        # Separate into heads
        q = self._separate_heads(q, self.num_heads)
        k = self._separate_heads(k, self.num_heads)
        v = self._separate_heads(v, self.num_heads)

        # Attention
        _, _, _, c_per_head = q.shape
        attn = q @ k.permute(0, 1, 3, 2)  # B x N_heads x N_tokens x N_tokens
        attn = attn / math.sqrt(c_per_head)
        attn = torch.softmax(attn, dim=-1)

        # Get output
        out = attn @ v
        out = self._recombine_heads(out)
        out = self.out_proj(out)

        return out



from .config import ModelConfig  # noqa: F401
from .embedding_model import EmbeddingModel  # noqa: F401
from .factory import create_model, create_preprocessing, transfer_weights  # noqa: F401
from .registry import (  # noqa: F401
    is_model,
    is_model_in_modules,
    is_model_pretrained,
    list_models,
    list_modules,
    model_class,
    model_config,
    register_model,
)
from .serialization import keras_serializable  # noqa: F401



# Copyright 2021 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PoolFormer implementation

Source: https://github.com/sail-sg/poolformer
"""
import copy
import os

import torch
import torch.nn as nn
from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.models.layers import DropPath, trunc_normal_
from timm.models.layers.helpers import to_2tuple
from timm.models.registry import register_model


def _cfg(url="", **kwargs):
    return {
        "url": url,
        "num_classes": 1000,
        "input_size": (3, 224, 224),
        "pool_size": None,
        "crop_pct": 0.95,
        "interpolation": "bicubic",
        "mean": IMAGENET_DEFAULT_MEAN,
        "std": IMAGENET_DEFAULT_STD,
        "classifier": "head",
        **kwargs,
    }


default_cfgs = {
    "poolformer_s": _cfg(crop_pct=0.9),
    "poolformer_m": _cfg(crop_pct=0.95),
}


class PatchEmbed(nn.Module):
    """
    Patch Embedding that is implemented by a layer of conv.
    Input: tensor in shape [B, C, H, W]
    Output: tensor in shape [B, C, H/stride, W/stride]
    """

    def __init__(
        self,
        patch_size=16,
        stride=16,
        padding=0,
        in_chans=3,
        embed_dim=768,
        norm_layer=None,
    ):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        stride = to_2tuple(stride)
        padding = to_2tuple(padding)
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=stride, padding=padding
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        x = self.proj(x)
        x = self.norm(x)
        return x


class LayerNormChannel(nn.Module):
    """
    LayerNorm only for Channel Dimension.
    Input: tensor in shape [B, C, H, W]
    """

    def __init__(self, num_channels, eps=1e-05):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(num_channels))
        self.bias = nn.Parameter(torch.zeros(num_channels))
        self.eps = eps

    def forward(self, x):
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        x = self.weight.unsqueeze(-1).unsqueeze(-1) * x + self.bias.unsqueeze(
            -1
        ).unsqueeze(-1)
        return x


class GroupNorm(nn.GroupNorm):
    """
    Group Normalization with 1 group.
    Input: tensor in shape [B, C, H, W]
    """

    def __init__(self, num_channels, **kwargs):
        super().__init__(1, num_channels, **kwargs)


class Pooling(nn.Module):
    """
    Implementation of pooling for PoolFormer
    --pool_size: pooling size
    """

    def __init__(self, pool_size=3):
        super().__init__()
        self.pool = nn.AvgPool2d(
            pool_size, stride=1, padding=pool_size // 2, count_include_pad=False
        )

    def forward(self, x):
        return self.pool(x) - x


class Mlp(nn.Module):
    """
    Implementation of MLP with 1*1 convolutions.
    Input: tensor with shape [B, C, H, W]
    """

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Conv2d(in_features, hidden_features, 1)
        self.act = act_layer()
        self.fc2 = nn.Conv2d(hidden_features, out_features, 1)
        self.drop = nn.Dropout(drop)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Conv2d):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class PoolFormerBlock(nn.Module):
    """
    Implementation of one PoolFormer block.
    --dim: embedding dim
    --pool_size: pooling size
    --mlp_ratio: mlp expansion ratio
    --act_layer: activation
    --norm_layer: normalization
    --drop: dropout rate
    --drop path: Stochastic Depth,
        refer to https://arxiv.org/abs/1603.09382
    --use_layer_scale, --layer_scale_init_value: LayerScale,
        refer to https://arxiv.org/abs/2103.17239
    """

    def __init__(
        self,
        dim,
        pool_size=3,
        mlp_ratio=4.0,
        act_layer=nn.GELU,
        norm_layer=GroupNorm,
        drop=0.0,
        drop_path=0.0,
        use_layer_scale=True,
        layer_scale_init_value=1e-5,
    ):

        super().__init__()

        self.norm1 = norm_layer(dim)
        self.token_mixer = Pooling(pool_size=pool_size)
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

        # The following two techniques are useful to train deep PoolFormers.
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.use_layer_scale = use_layer_scale
        if use_layer_scale:
            self.layer_scale_1 = nn.Parameter(
                layer_scale_init_value * torch.ones((dim)), requires_grad=True
            )
            self.layer_scale_2 = nn.Parameter(
                layer_scale_init_value * torch.ones((dim)), requires_grad=True
            )

    def forward(self, x):
        if self.use_layer_scale:
            x = x + self.drop_path(
                self.layer_scale_1.unsqueeze(-1).unsqueeze(-1)
                * self.token_mixer(self.norm1(x))
            )
            x = x + self.drop_path(
                self.layer_scale_2.unsqueeze(-1).unsqueeze(-1) * self.mlp(self.norm2(x))
            )
        else:
            x = x + self.drop_path(self.token_mixer(self.norm1(x)))
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


def basic_blocks(
    dim,
    index,
    layers,
    pool_size=3,
    mlp_ratio=4.0,
    act_layer=nn.GELU,
    norm_layer=GroupNorm,
    drop_rate=0.0,
    drop_path_rate=0.0,
    use_layer_scale=True,
    layer_scale_init_value=1e-5,
):
    """
    generate PoolFormer blocks for a stage
    return: PoolFormer blocks
    """
    blocks = []
    for block_idx in range(layers[index]):
        block_dpr = (
            drop_path_rate * (block_idx + sum(layers[:index])) / (sum(layers) - 1)
        )
        blocks.append(
            PoolFormerBlock(
                dim,
                pool_size=pool_size,
                mlp_ratio=mlp_ratio,
                act_layer=act_layer,
                norm_layer=norm_layer,
                drop=drop_rate,
                drop_path=block_dpr,
                use_layer_scale=use_layer_scale,
                layer_scale_init_value=layer_scale_init_value,
            )
        )
    blocks = nn.Sequential(*blocks)

    return blocks


class PoolFormer(nn.Module):
    """
    PoolFormer, the main class of our model
    --layers: [x,x,x,x], number of blocks for the 4 stages
    --embed_dims, --mlp_ratios, --pool_size: the embedding dims, mlp ratios and
        pooling size for the 4 stages
    --downsamples: flags to apply downsampling or not
    --norm_layer, --act_layer: define the types of normalizaiotn and activation
    --num_classes: number of classes for the image classification
    --in_patch_size, --in_stride, --in_pad: specify the patch embedding
        for the input image
    --down_patch_size --down_stride --down_pad:
        specify the downsample (patch embed.)
    --fork_faat: whetehr output features of the 4 stages, for dense prediction
    --init_cfg，--pretrained:
        for mmdetection and mmsegmentation to load pretrianfed weights
    """

    def __init__(
        self,
        layers,
        embed_dims=None,
        mlp_ratios=None,
        downsamples=None,
        pool_size=3,
        norm_layer=GroupNorm,
        act_layer=nn.GELU,
        num_classes=1000,
        in_patch_size=7,
        in_stride=4,
        in_pad=2,
        down_patch_size=3,
        down_stride=2,
        down_pad=1,
        drop_rate=0.0,
        drop_path_rate=0.0,
        use_layer_scale=True,
        layer_scale_init_value=1e-5,
        fork_feat=False,
        init_cfg=None,
        pretrained=None,
        **kwargs,
    ):

        super().__init__()

        if not fork_feat:
            self.num_classes = num_classes
        self.fork_feat = fork_feat

        self.patch_embed = PatchEmbed(
            patch_size=in_patch_size,
            stride=in_stride,
            padding=in_pad,
            in_chans=3,
            embed_dim=embed_dims[0],
        )

        # set the main block in network
        network = []
        for i in range(len(layers)):
            stage = basic_blocks(
                embed_dims[i],
                i,
                layers,
                pool_size=pool_size,
                mlp_ratio=mlp_ratios[i],
                act_layer=act_layer,
                norm_layer=norm_layer,
                drop_rate=drop_rate,
                drop_path_rate=drop_path_rate,
                use_layer_scale=use_layer_scale,
                layer_scale_init_value=layer_scale_init_value,
            )
            network.append(stage)
            if i >= len(layers) - 1:
                break
            if downsamples[i] or embed_dims[i] != embed_dims[i + 1]:
                # downsampling between two stages
                network.append(
                    PatchEmbed(
                        patch_size=down_patch_size,
                        stride=down_stride,
                        padding=down_pad,
                        in_chans=embed_dims[i],
                        embed_dim=embed_dims[i + 1],
                    )
                )

        self.network = nn.ModuleList(network)

        if self.fork_feat:
            # add a norm layer for each output
            self.out_indices = [0, 2, 4, 6]
            for i_emb, i_layer in enumerate(self.out_indices):
                if i_emb == 0 and os.environ.get("FORK_LAST3", None):
                    # TODOs: more elegant way
                    """For RetinaNet, `start_level=1`. The first norm layer will not used.
                    cmd: `FORK_LAST3=1 python -m torch.distributed.launch ...`
                    """
                    layer = nn.Identity()
                else:
                    layer = norm_layer(embed_dims[i_emb])
                layer_name = f"norm{i_layer}"
                self.add_module(layer_name, layer)
        else:
            # Classifier head
            self.norm = norm_layer(embed_dims[-1])
            self.head = (
                nn.Linear(embed_dims[-1], num_classes)
                if num_classes > 0
                else nn.Identity()
            )

        self.apply(self.cls_init_weights)

        self.init_cfg = copy.deepcopy(init_cfg)
        # load pre-trained model
        if self.fork_feat and (self.init_cfg is not None or pretrained is not None):
            self.init_weights()

    # init for classification
    def cls_init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes):
        self.num_classes = num_classes
        self.head = (
            nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()
        )

    def forward_embeddings(self, x):
        x = self.patch_embed(x)
        return x

    def forward_tokens(self, x):
        outs = []
        for idx, block in enumerate(self.network):
            x = block(x)
            if self.fork_feat and idx in self.out_indices:
                norm_layer = getattr(self, f"norm{idx}")
                x_out = norm_layer(x)
                outs.append(x_out)
        if self.fork_feat:
            # output the features of four stages for dense prediction
            return outs
        # output only the features of last layer for image classification
        return x

    def forward(self, x):
        # input embedding
        x = self.forward_embeddings(x)
        # through backbone
        x = self.forward_tokens(x)
        if self.fork_feat:
            # otuput features of four stages for dense prediction
            return x
        x = self.norm(x)
        cls_out = self.head(x.mean([-2, -1]))
        # for image classification
        return cls_out


@register_model
def poolformer_s12(pretrained=False, **kwargs):
    """
    PoolFormer-S12 model, Params: 12M
    --layers: [x,x,x,x], numbers of layers for the four stages
    --embed_dims, --mlp_ratios:
        embedding dims and mlp ratios for the four stages
    --downsamples: flags to apply downsampling or not in four blocks
    """
    layers = [2, 2, 6, 2]
    embed_dims = [64, 128, 320, 512]
    mlp_ratios = [4, 4, 4, 4]
    downsamples = [True, True, True, True]
    model = PoolFormer(
        layers,
        embed_dims=embed_dims,
        mlp_ratios=mlp_ratios,
        downsamples=downsamples,
        **kwargs,
    )
    model.default_cfg = default_cfgs["poolformer_s"]
    return model


@register_model
def poolformer_s24(pretrained=False, **kwargs):
    """
    PoolFormer-S24 model, Params: 21M
    """
    layers = [4, 4, 12, 4]
    embed_dims = [64, 128, 320, 512]
    mlp_ratios = [4, 4, 4, 4]
    downsamples = [True, True, True, True]
    model = PoolFormer(
        layers,
        embed_dims=embed_dims,
        mlp_ratios=mlp_ratios,
        downsamples=downsamples,
        **kwargs,
    )
    model.default_cfg = default_cfgs["poolformer_s"]
    return model


@register_model
def poolformer_s36(pretrained=False, **kwargs):
    """
    PoolFormer-S36 model, Params: 31M
    """
    layers = [6, 6, 18, 6]
    embed_dims = [64, 128, 320, 512]
    mlp_ratios = [4, 4, 4, 4]
    downsamples = [True, True, True, True]
    model = PoolFormer(
        layers,
        embed_dims=embed_dims,
        mlp_ratios=mlp_ratios,
        downsamples=downsamples,
        layer_scale_init_value=1e-6,
        **kwargs,
    )
    model.default_cfg = default_cfgs["poolformer_s"]
    return model


@register_model
def poolformer_m36(pretrained=False, **kwargs):
    """
    PoolFormer-M36 model, Params: 56M
    """
    layers = [6, 6, 18, 6]
    embed_dims = [96, 192, 384, 768]
    mlp_ratios = [4, 4, 4, 4]
    downsamples = [True, True, True, True]
    model = PoolFormer(
        layers,
        embed_dims=embed_dims,
        mlp_ratios=mlp_ratios,
        downsamples=downsamples,
        layer_scale_init_value=1e-6,
        **kwargs,
    )
    model.default_cfg = default_cfgs["poolformer_m"]
    return model


@register_model
def poolformer_m48(pretrained=False, **kwargs):
    """
    PoolFormer-M48 model, Params: 73M
    """
    layers = [8, 8, 24, 8]
    embed_dims = [96, 192, 384, 768]
    mlp_ratios = [4, 4, 4, 4]
    downsamples = [True, True, True, True]
    model = PoolFormer(
        layers,
        embed_dims=embed_dims,
        mlp_ratios=mlp_ratios,
        downsamples=downsamples,
        layer_scale_init_value=1e-6,
        **kwargs,
    )
    model.default_cfg = default_cfgs["poolformer_m"]
    return model



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from typing import Tuple

import torch
import torch.nn as nn
from torch.nn import functional as F

from ..modeling import Sam
from .amg import calculate_stability_score


class SamOnnxModel(nn.Module):
    """
    This model should not be called directly, but is used in ONNX export.
    It combines the prompt encoder, mask decoder, and mask postprocessing of Sam,
    with some functions modified to enable model tracing. Also supports extra
    options controlling what information. See the ONNX export script for details.
    """

    def __init__(
        self,
        model: Sam,
        return_single_mask: bool,
        use_stability_score: bool = False,
        return_extra_metrics: bool = False,
    ) -> None:
        super().__init__()
        self.mask_decoder = model.mask_decoder
        self.model = model
        self.img_size = model.image_encoder.img_size
        self.return_single_mask = return_single_mask
        self.use_stability_score = use_stability_score
        self.stability_score_offset = 1.0
        self.return_extra_metrics = return_extra_metrics

    @staticmethod
    def resize_longest_image_size(
        input_image_size: torch.Tensor, longest_side: int
    ) -> torch.Tensor:
        input_image_size = input_image_size.to(torch.float32)
        scale = longest_side / torch.max(input_image_size)
        transformed_size = scale * input_image_size
        transformed_size = torch.floor(transformed_size + 0.5).to(torch.int64)
        return transformed_size

    def _embed_points(
        self, point_coords: torch.Tensor, point_labels: torch.Tensor
    ) -> torch.Tensor:
        point_coords = point_coords + 0.5
        point_coords = point_coords / self.img_size
        point_embedding = self.model.prompt_encoder.pe_layer._pe_encoding(point_coords)
        point_labels = point_labels.unsqueeze(-1).expand_as(point_embedding)

        point_embedding = point_embedding * (point_labels != -1)
        point_embedding = (
            point_embedding
            + self.model.prompt_encoder.not_a_point_embed.weight * (point_labels == -1)
        )

        for i in range(self.model.prompt_encoder.num_point_embeddings):
            point_embedding = (
                point_embedding
                + self.model.prompt_encoder.point_embeddings[i].weight
                * (point_labels == i)
            )

        return point_embedding

    def _embed_masks(
        self, input_mask: torch.Tensor, has_mask_input: torch.Tensor
    ) -> torch.Tensor:
        mask_embedding = has_mask_input * self.model.prompt_encoder.mask_downscaling(
            input_mask
        )
        mask_embedding = mask_embedding + (
            1 - has_mask_input
        ) * self.model.prompt_encoder.no_mask_embed.weight.reshape(1, -1, 1, 1)
        return mask_embedding

    def mask_postprocessing(
        self, masks: torch.Tensor, orig_im_size: torch.Tensor
    ) -> torch.Tensor:
        masks = F.interpolate(
            masks,
            size=(self.img_size, self.img_size),
            mode="bilinear",
            align_corners=False,
        )

        prepadded_size = self.resize_longest_image_size(orig_im_size, self.img_size)
        masks = masks[..., : int(prepadded_size[0]), : int(prepadded_size[1])]

        orig_im_size = orig_im_size.to(torch.int64)
        h, w = orig_im_size[0], orig_im_size[1]
        masks = F.interpolate(masks, size=(h, w), mode="bilinear", align_corners=False)
        return masks

    def select_masks(
        self, masks: torch.Tensor, iou_preds: torch.Tensor, num_points: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # Determine if we should return the multiclick mask or not from the number of
        # points. The reweighting is used to avoid control flow.
        score_reweight = torch.tensor(
            [[1000] + [0] * (self.model.mask_decoder.num_mask_tokens - 1)]
        ).to(iou_preds.device)
        score = iou_preds + (num_points - 2.5) * score_reweight
        best_idx = torch.argmax(score, dim=1)
        masks = masks[torch.arange(masks.shape[0]), best_idx, :, :].unsqueeze(1)
        iou_preds = iou_preds[torch.arange(masks.shape[0]), best_idx].unsqueeze(1)

        return masks, iou_preds

    @torch.no_grad()
    def forward(
        self,
        image_embeddings: torch.Tensor,
        point_coords: torch.Tensor,
        point_labels: torch.Tensor,
        mask_input: torch.Tensor,
        has_mask_input: torch.Tensor,
        orig_im_size: torch.Tensor,
    ):
        sparse_embedding = self._embed_points(point_coords, point_labels)
        dense_embedding = self._embed_masks(mask_input, has_mask_input)

        masks, scores = self.model.mask_decoder.predict_masks(
            image_embeddings=image_embeddings,
            image_pe=self.model.prompt_encoder.get_dense_pe(),
            sparse_prompt_embeddings=sparse_embedding,
            dense_prompt_embeddings=dense_embedding,
        )

        if self.use_stability_score:
            scores = calculate_stability_score(
                masks, self.model.mask_threshold, self.stability_score_offset
            )

        if self.return_single_mask:
            masks, scores = self.select_masks(masks, scores, point_coords.shape[1])

        upscaled_masks = self.mask_postprocessing(masks, orig_im_size)

        if self.return_extra_metrics:
            stability_scores = calculate_stability_score(
                upscaled_masks, self.model.mask_threshold, self.stability_score_offset
            )
            areas = (upscaled_masks > self.model.mask_threshold).sum(-1).sum(-1)
            return upscaled_masks, scores, stability_scores, areas, masks

        return upscaled_masks, scores, masks



"""
MLP-Mixer, ResMLP, and gMLP models

This implementation is ported from the timm, which is based on the original
implementation from the MLP-Mixer paper.

Official JAX impl:
https://github.com/google-research/vision_transformer/blob/linen/vit_jax/models_mixer.py

Paper: MLP-Mixer: An all-MLP Architecture for Vision
Arxiv: https://arxiv.org/abs/2105.01601

Also supporting ResMLP, and a preliminary implementation of gMLP

Code: https://github.com/facebookresearch/deit
Paper: ResMLP: Feedforward networks for image classification...
Arxiv: https://arxiv.org/abs/2105.03404

Paper: Pay Attention to MLPs
Arxiv: https://arxiv.org/abs/2105.08050

A thank you to paper authors for releasing code and weights.

Copyright 2021 Martins Bruveris
Copyright 2021 Ross Wightman
"""
from dataclasses import dataclass
from typing import List, Tuple

import tensorflow as tf

from tfimm.layers import (
    MLP,
    DropPath,
    GatedMLP,
    GluMLP,
    PatchEmbeddings,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint function to this
__all__ = ["MLPMixer", "MLPMixerConfig"]


@dataclass
class MLPMixerConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: int = 16
    embed_dim: int = 512
    nb_blocks: int = 16
    mlp_ratio: Tuple[float, float] = (0.5, 4.0)
    block_layer: str = "mixer_block"
    mlp_layer: str = "mlp"
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    init_values: float = 1e-4  # Initialisation for ResBlocks
    nlhb: bool = False  # Negative logarithmic head bias
    stem_norm: bool = False
    # Parameters for inference
    crop_pct: float = 0.875
    interpolation: str = "bicubic"
    mean: Tuple[float, float, float] = (0.5, 0.5, 0.5)
    std: Tuple[float, float, float] = (0.5, 0.5, 0.5)
    # Weight transfer
    first_conv: str = "stem/proj"
    classifier: str = "head"

    @property
    def nb_patches(self) -> int:
        return (self.input_size[0] // self.patch_size) * (
            self.input_size[1] // self.patch_size
        )


class MixerBlock(tf.keras.layers.Layer):
    """
    Residual Block w/ token mixing and channel MLPs
    Based on: "MLP-Mixer: An all-MLP Architecture for Vision"
    """

    def __init__(self, cfg: MLPMixerConfig, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        mlp_layer = MLP_LAYER_DICT[cfg.mlp_layer]
        tokens_dim, channels_dim = [int(x * cfg.embed_dim) for x in cfg.mlp_ratio]

        self.norm1 = norm_layer(name="norm1")
        self.mlp_tokens = mlp_layer(
            hidden_dim=tokens_dim,
            embed_dim=cfg.nb_patches,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp_tokens",
        )
        self.drop_path = DropPath(drop_prob=cfg.drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp_channels = mlp_layer(
            hidden_dim=channels_dim,
            embed_dim=cfg.embed_dim,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp_channels",
        )

    def call(self, x, training=False):
        shortcut = x
        x = self.norm1(x, training=training)
        x = tf.transpose(x, perm=(0, 2, 1))
        x = self.mlp_tokens(x, training=training)
        x = tf.transpose(x, perm=(0, 2, 1))
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp_channels(x, training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut

        return x


class ResBlock(tf.keras.layers.Layer):
    """
    Residual MLP block with LayerScale

    Based on: ResMLP: Feedforward networks for image classification...
    """

    def __init__(self, cfg: MLPMixerConfig, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        mlp_layer = MLP_LAYER_DICT[cfg.mlp_layer]
        self.norm1 = norm_layer(name="norm1")
        self.linear_tokens = tf.keras.layers.Dense(
            units=cfg.nb_patches,
            name="linear_tokens",
        )
        self.drop_path = DropPath(drop_prob=cfg.drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp_channels = mlp_layer(
            hidden_dim=int(cfg.embed_dim * cfg.mlp_ratio[1]),
            embed_dim=cfg.embed_dim,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp_channels",
        )

    def build(self, input_shape):
        self.ls1 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(self.cfg.init_values),
            trainable=True,
            name="ls1",
        )
        self.ls2 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(self.cfg.init_values),
            trainable=True,
            name="ls2",
        )

    def call(self, x, training=False):
        shortcut = x
        x = self.norm1(x, training=training)
        x = tf.transpose(x, perm=(0, 2, 1))
        x = self.linear_tokens(x, training=training)
        x = tf.transpose(x, perm=(0, 2, 1))
        x = self.ls1 * x
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp_channels(x, training=training)
        x = self.ls2 * x
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


class SpatialGatingBlock(tf.keras.layers.Layer):
    """
    Residual Block with Spatial Gating

    Based on: Pay Attention to MLPs - https://arxiv.org/abs/2105.08050
    """

    def __init__(self, cfg: MLPMixerConfig, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        mlp_layer = MLP_LAYER_DICT[cfg.mlp_layer]
        self.norm = norm_layer(name="norm")
        self.mlp_channels = mlp_layer(
            hidden_dim=int(cfg.embed_dim * cfg.mlp_ratio[1]),
            embed_dim=cfg.embed_dim,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp_channels",
        )
        self.drop_path = DropPath(drop_prob=cfg.drop_path_rate)

    def call(self, x, training=False):
        shortcut = x
        x = self.norm(x, training=training)
        x = self.mlp_channels(x, training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


BLOCK_LAYER_DICT = {
    "mixer_block": MixerBlock,
    "res_block": ResBlock,
    "spatial_gating_block": SpatialGatingBlock,
}

MLP_LAYER_DICT = {
    "mlp": MLP,
    "glu_mlp": GluMLP,
    "gated_mlp": GatedMLP,
}


@keras_serializable
class MLPMixer(tf.keras.Model):
    cfg_class = MLPMixerConfig

    def __init__(self, cfg: MLPMixerConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg
        self.nb_features = cfg.embed_dim

        norm_layer = norm_layer_factory(cfg.norm_layer)
        block_layer = BLOCK_LAYER_DICT[cfg.block_layer]

        self.stem = PatchEmbeddings(
            patch_size=cfg.patch_size,
            embed_dim=cfg.embed_dim,
            norm_layer=cfg.norm_layer if cfg.stem_norm else "",
            name="stem",
        )
        self.blocks = [
            block_layer(cfg=cfg, name=f"blocks/{j}") for j in range(cfg.nb_blocks)
        ]
        self.norm = norm_layer(name="norm")
        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        return (
            ["stem"]
            + [f"block_{j}" for j in range(self.cfg.nb_blocks)]
            + ["features_all", "features", "logits"]
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x = self.stem(x, training=training)
        features["stem"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x

        x = self.norm(x, training=training)
        features["features_all"] = x

        x = tf.reduce_mean(x, axis=1)
        features["features"] = x

        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def mixer_s32_224():
    """
    Mixer-S/32 224x224
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_s32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=512,
        nb_blocks=8,
    )
    return MLPMixer, cfg


@register_model
def mixer_s16_224():
    """
    Mixer-S/16 224x224
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_s16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=512,
        nb_blocks=8,
    )
    return MLPMixer, cfg


@register_model
def mixer_b32_224():
    """
    Mixer-B/32 224x224
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_b32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=768,
        nb_blocks=12,
    )
    return MLPMixer, cfg


@register_model
def mixer_b16_224():
    """Mixer-B/16 224x224. ImageNet-1k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_b16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
    )
    return MLPMixer, cfg


@register_model
def mixer_b16_224_in21k():
    """
    Mixer-B/16 224x224. ImageNet-21k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_b16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
    )
    return MLPMixer, cfg


@register_model
def mixer_l32_224():
    """
    Mixer-L/32 224x224.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_l32_224",
        url="[timm]",
        patch_size=32,
        embed_dim=1024,
        nb_blocks=24,
    )
    return MLPMixer, cfg


@register_model
def mixer_l16_224():
    """
    Mixer-L/16 224x224. ImageNet-1k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_l16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=1024,
        nb_blocks=24,
    )
    return MLPMixer, cfg


@register_model
def mixer_l16_224_in21k():
    """
    Mixer-L/16 224x224. ImageNet-21k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    """
    cfg = MLPMixerConfig(
        name="mixer_l16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        patch_size=16,
        embed_dim=1024,
        nb_blocks=24,
    )
    return MLPMixer, cfg


@register_model
def mixer_b16_224_miil():
    """
    Mixer-B/16 224x224. ImageNet-21k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    """
    cfg = MLPMixerConfig(
        name="mixer_b16_224_miil",
        url="[timm]",
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        crop_pct=0.875,
        interpolation="bilinear",
        mean=(0.0, 0.0, 0.0),
        std=(1.0, 1.0, 1.0),
    )
    return MLPMixer, cfg


@register_model
def mixer_b16_224_miil_in21k():
    """
    Mixer-B/16 224x224. ImageNet-1k pretrained weights.
    Paper: MLP-Mixer: An all-MLP Architecture for Vision
    Link: https://arxiv.org/abs/2105.01601
    Weights taken from: https://github.com/Alibaba-MIIL/ImageNet21K
    """
    cfg = MLPMixerConfig(
        name="mixer_b16_224_miil_in21k",
        url="[timm]",
        nb_classes=11221,
        patch_size=16,
        embed_dim=768,
        nb_blocks=12,
        crop_pct=0.875,
        interpolation="bilinear",
        mean=(0.0, 0.0, 0.0),
        std=(1.0, 1.0, 1.0),
    )
    return MLPMixer, cfg


@register_model
def gmixer_12_224():
    """
    Glu-Mixer-12 224x224
    Experiment by Ross Wightman, adding (Si)GLU to MLP-Mixer
    """
    cfg = MLPMixerConfig(
        name="gmixer_12_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        mlp_ratio=(1.0, 4.0),
        mlp_layer="glu_mlp",
        act_layer="swish",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def gmixer_24_224():
    """
    Glu-Mixer-24 224x224
    Experiment by Ross Wightman, adding (Si)GLU to MLP-Mixer
    """
    cfg = MLPMixerConfig(
        name="gmixer_24_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        mlp_ratio=(1.0, 4.0),
        mlp_layer="glu_mlp",
        act_layer="swish",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_12_224():
    """
    ResMLP-12
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_12_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_24_224():
    """
    ResMLP-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_24_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-5,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_36_224():
    """
    ResMLP-36
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_36_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=36,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-6,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_big_24_224():
    """
    ResMLP-B-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_big_24_224",
        url="[timm]",
        patch_size=8,
        embed_dim=768,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-6,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_12_distilled_224():
    """
    ResMLP-12
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_12_distilled_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_24_distilled_224():
    """
    ResMLP-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_24_distilled_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-5,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_36_distilled_224():
    """
    ResMLP-36
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_36_distilled_224",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=36,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-6,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_big_24_distilled_224():
    """
    ResMLP-B-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_big_24_distilled_224",
        url="[timm]",
        patch_size=8,
        embed_dim=768,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-6,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_big_24_224_in22ft1k():
    """
    ResMLP-B-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    """
    cfg = MLPMixerConfig(
        name="resmlp_big_24_224_in22ft1k",
        url="[timm]",
        patch_size=8,
        embed_dim=768,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-6,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_12_224_dino():
    """
    ResMLP-12
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    Model pretrained via DINO (self-supervised) - https://arxiv.org/abs/2104.14294
    """
    cfg = MLPMixerConfig(
        name="resmlp_12_224_dino",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=12,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def resmlp_24_224_dino():
    """
    ResMLP-24
    Paper: ResMLP: Feedforward networks for image classification...
    Link: https://arxiv.org/abs/2105.03404
    Model pretrained via DINO (self-supervised) - https://arxiv.org/abs/2104.14294
    """
    cfg = MLPMixerConfig(
        name="resmlp_24_224_dino",
        url="[timm]",
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        mlp_ratio=(4.0, 4.0),
        block_layer="res_block",
        init_values=1e-5,
        norm_layer="affine",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    )
    return MLPMixer, cfg


@register_model
def gmlp_ti16_224():
    """
    gMLP-Tiny
    Paper: Pay Attention to MLPs
    Link: https://arxiv.org/abs/2105.08050
    """
    cfg = MLPMixerConfig(
        name="gmlp_ti16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=128,
        nb_blocks=30,
        mlp_ratio=(6.0, 6.0),
        block_layer="spatial_gating_block",
        mlp_layer="gated_mlp",
    )
    return MLPMixer, cfg


@register_model
def gmlp_s16_224():
    """
    gMLP-Small
    Paper: Pay Attention to MLPs
    Link: https://arxiv.org/abs/2105.08050
    """
    cfg = MLPMixerConfig(
        name="gmlp_s16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=256,
        nb_blocks=30,
        mlp_ratio=(6.0, 6.0),
        block_layer="spatial_gating_block",
        mlp_layer="gated_mlp",
    )
    return MLPMixer, cfg


@register_model
def gmlp_b16_224():
    """
    gMLP-Base
    Paper: Pay Attention to MLPs
    Link: https://arxiv.org/abs/2105.08050
    """
    cfg = MLPMixerConfig(
        name="gmlp_b16_224",
        url="[timm]",
        patch_size=16,
        embed_dim=512,
        nb_blocks=30,
        mlp_ratio=(6.0, 6.0),
        block_layer="spatial_gating_block",
        mlp_layer="gated_mlp",
    )
    return MLPMixer, cfg



from .cait import *  # noqa: F401
from .convmixer import *  # noqa: F401
from .convnext import *  # noqa: F401
from .efficientnet import *  # noqa: F401
from .mlp_mixer import *  # noqa: F401
from .pit import *  # noqa: F401
from .poolformer import *  # noqa: F401
from .pvt import *  # noqa: F401
from .pvt_v2 import *  # noqa: F401
from .resnet import *  # noqa: F401
from .resnetv2 import *  # noqa: F401
from .swin import *  # noqa: F401
from .vgg import *  # noqa: F401
from .vit import *  # noqa: F401
from .vit_hybrid import *  # noqa: F401



"""
Attention layers used in CNNs
"""
import math

import tensorflow as tf

from ..utils import make_divisible
from .factory import act_layer_factory, norm_layer_factory


class SEModule(tf.keras.layers.Layer):
    """
    SE Module as defined in original SE-Nets with a few additions

    Paper: Squeeze-and-Excitation Networks
    Link: https://arxiv.org/abs/1709.01507

    Additions include:
        * divisor can be specified to keep channels % div == 0 (default: 8)
        * reduction channels can be specified directly by arg (if rd_channels is set)
        * reduction channels can be specified by float rd_ratio (default: 1/16)
        * global max pooling can be added to the squeeze aggregation
        * customizable activation, normalization, and gate layer
    """

    def __init__(
        self,
        rd_ratio=1.0 / 16,
        rd_channels=None,
        rd_divisor=8,
        act_layer="relu",
        norm_layer="",
        gate_layer="sigmoid",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.rd_ratio = rd_ratio
        self.rd_channels = rd_channels
        self.rd_divisor = rd_divisor
        act_layer = act_layer_factory(act_layer)
        norm_layer = norm_layer_factory(norm_layer)
        gate_layer = act_layer_factory(gate_layer)

        self.bn = norm_layer(name="bn")
        self.act = act_layer()
        self.gate = gate_layer()

    def build(self, input_shape):
        channels = input_shape[-1]
        rd_channels = self.rd_channels or make_divisible(
            channels * self.rd_ratio, self.rd_divisor, round_limit=0.0
        )
        self.fc1 = tf.keras.layers.Conv2D(
            filters=rd_channels,
            kernel_size=1,
            use_bias=True,
            name="fc1",
        )
        self.fc2 = tf.keras.layers.Conv2D(
            filters=channels,
            kernel_size=1,
            use_bias=True,
            name="fc2",
        )

    def call(self, x, training=False):
        x_se = tf.reduce_mean(x, axis=(1, 2), keepdims=True)
        x_se = self.fc1(x_se)
        x_se = self.bn(x_se, training=training)
        x_se = self.act(x_se)
        x_se = self.fc2(x_se)
        x_se = self.gate(x_se)
        x = x * x_se
        return x


class EcaModule(tf.keras.layers.Layer):
    """Constructs an ECA module.

    Paper: ECA-Net: Efficient Channel Attention for Deep Convolutional Neural Networks
    Link: https://arxiv.org/pdf/1910.03151.pdf

    Args:
        gamma: Used in kernel_size calc, see original paper for details
            refer to original paper https://arxiv.org/pdf/1910.03151.pdf
        beta: Used in kernel_size calc, see above
        gate_layer: Gating non-linearity to use
    """

    def __init__(
        self,
        gamma: int = 2,
        beta: int = 1,
        gate_layer: str = "sigmoid",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.gamma = gamma
        self.beta = beta

        gate_layer = act_layer_factory(gate_layer)
        self.gate = gate_layer()

    def build(self, input_shape):
        # Kernel size is determined dynamically from number of input channels
        channels = input_shape[-1]
        t = int(abs(math.log(channels, 2) + self.beta) / self.gamma)
        kernel_size = max(t if t % 2 else t + 1, 3)
        padding = (kernel_size - 1) // 2

        self.pad = tf.keras.layers.ZeroPadding1D(padding=padding)
        self.conv = tf.keras.layers.Conv1D(
            filters=1,
            kernel_size=kernel_size,
            use_bias=False,
            name="conv",
        )

    def call(self, x, training=False):
        # x is (N, H, W, C)
        y = tf.reduce_mean(x, axis=(1, 2))  # (N, C)
        y = tf.expand_dims(y, axis=-1)  # (N, C, 1)
        y = self.pad(y)
        y = self.conv(y)  # (N, C, 1)
        y = self.gate(y)
        y = tf.transpose(y, perm=(0, 2, 1))  # (N, 1, C)
        y = tf.expand_dims(y, axis=1)  # (N, 1, 1, C)
        x = y * x
        return x


def attn_layer_factory(attn_layer: str):
    """Returns a function that creates the required attention layer."""
    if attn_layer == "":
        return lambda **kwargs: tf.keras.layers.Activation("linear")
    # Lightweight attention modules (channel and/or coarse spatial).
    # Typically, added to existing network architecture blocks in addition to existing
    # convolutions.
    elif attn_layer == "eca":
        return lambda **kwargs: EcaModule(**kwargs)
    elif attn_layer == "se":
        return lambda **kwargs: SEModule(**kwargs)
    else:
        raise ValueError(f"Unknown attention layer: {attn_layer}.")



import numpy as np
import tensorflow as tf


class BlurPool2D(tf.keras.layers.Layer):
    """
    BlurPool layer as described in the paper "Making Convolutional Networks
    Shift-Invariant Again" by Zhang et al.

    Link: https://arxiv.org/pdf/1904.11486.pdf

    This implementation is adapted from:
    https://github.com/csvance/blur-pool-keras/blob/master/blurpool.py
    """

    def __init__(self, kernel_size: int = 3, stride: int = 2, **kwargs):
        super().__init__(**kwargs)

        self.kernel_size = kernel_size
        self.stride = stride
        p = (kernel_size + stride) // 2 - 1
        self.paddings = [[0, 0], [p, p], [p, p], [0, 0]]
        self.blur_kernel = None

    def build(self, input_shape):
        if self.kernel_size == 3:
            bk = np.array([[1, 2, 1], [2, 4, 2], [1, 2, 1]])
            bk = bk / np.sum(bk)
        elif self.kernel_size == 5:
            bk = np.array(
                [
                    [1, 4, 6, 4, 1],
                    [4, 16, 24, 16, 4],
                    [6, 24, 36, 24, 6],
                    [4, 16, 24, 16, 4],
                    [1, 4, 6, 4, 1],
                ]
            )
            bk = bk / np.sum(bk)
        else:
            raise ValueError(f"Kernel size {self.kernel_size} not implemented.")

        channels = input_shape[-1]
        bk = np.repeat(bk, channels)
        bk = np.reshape(bk, (self.kernel_size, self.kernel_size, channels, 1))

        self.blur_kernel = self.add_weight(
            name="blur_kernel",
            shape=(self.kernel_size, self.kernel_size, channels, 1),
            initializer=tf.keras.initializers.constant(bk),
            trainable=False,
        )

    def call(self, x):
        x = tf.pad(x, paddings=self.paddings, mode="REFLECT")
        x = tf.nn.depthwise_conv2d(
            input=x,
            filter=self.blur_kernel,
            strides=[1, self.stride, self.stride, 1],
            padding="VALID",
        )
        return x

    def cast_inputs(self, inputs):
        # Casts to float16, the policy's lowest-precision dtype
        return self._mixed_precision_policy.cast_to_lowest(inputs)



import argparse
import ast
import copy
import dataclasses
import logging
import sys
from pathlib import Path

import yaml

from .registry import get_cfg_class


def to_dict_format(cfg):
    """
    A configuration is a nested dictionary with potentially dataclasses as values. This
    function converts dataclasses to dictionaries for easier serialization.
    """
    if dataclasses.is_dataclass(cfg):
        cfg = dataclasses.asdict(cfg)
        return to_dict_format(cfg)

    res_cfg = {}
    for key, val in cfg.items():
        if isinstance(val, dict):
            res_cfg[key] = to_dict_format(val)
        elif dataclasses.is_dataclass(val):
            res_cfg[key] = dataclasses.asdict(val)
        else:
            res_cfg[key] = val
    return res_cfg


def to_cls_format(cfg):
    """
    Converts a configuration in dictionary format to config classes, using the type
    information provided in "_class" fields.
    """
    res_cfg = {}
    for key, val in cfg.items():
        if isinstance(val, dict):
            # First recurse into dictionary before converting to class
            val = to_cls_format(val)

            cls = cfg[f"{key}_class"]
            if cls:  # Now convert to correct class, if class is set
                cls = get_cfg_class(cls)
                res_cfg[key] = cls(**val)
            else:
                res_cfg[key] = None
        else:
            res_cfg[key] = val
    return res_cfg


def to_arg_format(cfg):
    """
    Converts the values of a config in dictionary format to tuples (type, value). This
    is needed to separate dtype from default value for argument parsing.

    We treat `None` and `MISSING` as `str`, which is a useful default for parsing
    command line arguments.

    Example:
    ```
    >>> cfg = {"a": 3, "b": {"c": "s"}}
    >>> print(to_arg_format(cfg))
    {"a": (int, 3), "b": {"c": (str, "s")}}
    ```
    """
    res_cfg = {}
    for key, val in cfg.items():
        if isinstance(val, dict):
            res_cfg[key] = to_arg_format(val)
        else:
            tp = type(val) if val not in {None, dataclasses.MISSING} else str
            res_cfg[key] = (tp, val)
    return res_cfg


def check_format(cfg):
    """
    We apply the following checks to configurations in dictionary format:
     - For fields ending in `_class`, map None to empty string.
     - If the `xyz_class` field is present, nesting has to happen and we add the `xyz`
       field and initialize it with an empty dictionary.
     - Nesting is only allowed, if the `_class` field is present, i.e., if `cfg[key]`
       is a dictionary, then `cfg[key + "_class"]` has to exist.
     - If the keys `xyz` and `xyz_class` both exist, then `xyz` has to be a dictionary.
       None is mapped to the empty dictionary. An exception is raised for any other
       values.
    """
    res_cfg = {}
    for key, val in cfg.items():
        if key.endswith("_class"):
            if not isinstance(val, str) and val is not None:
                raise ValueError(f"Value for key {key} should be a string.")
            res_cfg[key] = "" if val is None else val

            stem = key[: -len("_class")]  # Remove suffix
            if stem not in cfg:
                res_cfg[stem] = {}
        elif isinstance(val, dict):
            if key + "_class" not in cfg:
                raise ValueError(f"Nesting only allowed if key `{key}_class` exists.")
            res_cfg[key] = check_format(val)
        else:
            if key + "_class" in cfg:
                if val is not None:
                    raise ValueError(f"Value for key {key} has to be a dict.")
                res_cfg[key] = {}
            else:
                res_cfg[key] = val
    return res_cfg


def add_default_args(cfg):
    """
    When `cfg` is a config in dictionary format with values of the form (type, value),
    this function adds default fields from the configuration dataclasses specified in
    "_class" fields.

    If `cfg = {"data_class": "Foo"}`, the function will look up the config class
    associated to `Foo` and add all fields of the config class as nested dictionary
    `cfg["data"] = {...}`.
    """
    res_cfg = {}
    for key, val in cfg.items():
        if key.endswith("_class"):
            res_cfg[key] = val  # We need to copy the field itself
            if val[1] == "":  # The class has not yet been specified, do nothing.
                continue

            cls = get_cfg_class(val[1])
            stem = key[: -len("_class")]  # Remove suffix
            fields = dataclasses.fields(cls)
            # Fields with missing default value will be assigned value MISSING
            params = {field.name: (field.type, field.default) for field in fields}
            if stem in cfg:
                if not isinstance(cfg[stem], dict):
                    raise ValueError(
                        f"cfg[{stem}] should be a dict, but is {type(cfg[stem])}."
                    )
                # If some fields are already set in the config, we use those values,
                # i.e., whatever is in the config has higher priority over default
                # values.
                for sub_key, sub_val in cfg[stem].items():
                    if sub_key in params:
                        params[sub_key] = sub_val
            # Recursively resolve subclasses
            res_cfg[stem] = add_default_args(params)
        elif isinstance(val, dict) and key + "_class" not in cfg:
            # We need to check for key `xyz` if `xyz_class` is not a key, because
            # otherwise we don't know whether we process key `xyz` first or `xyz_class`.
            res_cfg[key] = add_default_args(val)
        elif key + "_class" not in cfg:
            res_cfg[key] = val
    return res_cfg


def add_cls_default_args(cfg, cls):
    """
    Function adds all fields of a config class to a config in argument format.
    """
    res_cfg = copy.deepcopy(cfg)
    if cls is None:
        return res_cfg
    fields = dataclasses.fields(cls)
    params = {field.name: (field.type, field.default) for field in fields}
    # If some fields are already set in the config, we use those values, i.e., whatever
    # is in the config has higher priority over default values.
    for key, val in params.items():
        # We need to ignore nested fields, since those are not parsed directly.
        if key + "_class" in params:
            continue
        if key not in res_cfg:
            res_cfg[key] = val
    return res_cfg


def add_cfg_file_arg(cfg, cls):
    """
    If `cls` has a `cfg_file` field, we add this field to `cfg`.
    """
    cfg = copy.deepcopy(cfg)
    if "cfg_file" in cfg or cls is None:
        return cfg

    fields = dataclasses.fields(cls)
    params = {field.name: field.default for field in fields}
    if "cfg_file" in params:
        val = params["cfg_file"]
        cfg["cfg_file"] = val if val != dataclasses.MISSING else ""
    return cfg


def deep_to_flat(cfg):
    """
    Function flattens a nested config in dictionary format by joining keys of nested
    dictionaries with ".".

    For example,
    ```
    >>> cfg = {"a": {"b": 1, "c": 2}, "d": 3}
    >>> print(deep_to_flat(cfg))
    {"a.b": 1, "a.c": 2, "d": 3}
    ```
    """
    res_cfg = {}
    for key, val in cfg.items():
        if isinstance(val, dict):
            val = deep_to_flat(val)
            # After recursive flattening, `val` is now a flat dictionary. Add its keys
            # to the current level
            for sub_key, sub_val in val.items():
                res_cfg[f"{key}.{sub_key}"] = sub_val
        else:
            res_cfg[key] = val
    return res_cfg


def flat_to_deep(cfg):
    """
    Function converts a flat config to a nested config in dictionary format. This is
    the inverse of `deep_to_flat()`.
    """
    res_cfg = {}
    # By iterating over the items of `cfg` we resolve one level of nesting.
    for key, val in cfg.items():
        if "." in key:
            root, leaf = tuple(key.split(".", 1))  # Split off the first part
            if root not in res_cfg:
                res_cfg[root] = {}
            res_cfg[root][leaf] = val
        else:
            res_cfg[key] = val

    # Now we iterate again and call the function recursively to resolve deeper levels
    # of nesting.
    for key, val in res_cfg.items():
        if isinstance(val, dict):
            res_cfg[key] = flat_to_deep(val)

    return res_cfg


def dump_config(cfg, filename):
    """Converts a config to nested dictionaries and saves them in yaml format."""
    cfg = to_dict_format(cfg)
    # Create directory for `filename` if it doesn't exist
    Path(filename).parents[0].mkdir(parents=True, exist_ok=True)
    with open(filename, "w") as yaml_file:
        yaml.dump(cfg, yaml_file, default_flow_style=False, sort_keys=False)


def apply_cfg_file(cfg, args):
    """
    We check if a config file is expected (presence of `cfg_file` key in `cfg`). If so,
    see, if an updated config file is passed via command line. Then we read the config
    file and apply the values to the config.
    """
    # First we need to check if the user has supplied a config file
    parser = argparse.ArgumentParser()
    parser.add_argument("--cfg_file", default=cfg["cfg_file"])
    namespace, _ = parser.parse_known_args(args)
    cfg_file = namespace.cfg_file

    # No config file is present
    if not cfg_file:
        return cfg

    with open(cfg_file, "r") as f:
        loaded_cfg = yaml.load(f, Loader=yaml.Loader)

    # Updating is easiest done with flattened configs
    cfg = deep_to_flat(cfg)
    loaded_cfg = deep_to_flat(loaded_cfg)
    cfg.update(loaded_cfg)  # Because of deep_to_flat() we are operating on a copy

    # And now back to nested ones
    cfg = flat_to_deep(cfg)
    cfg = check_format(cfg)

    return cfg


def str2bool(v):
    """
    Converts various string to bool, accepting various represenations of True/False.
    """
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def arg_as_tuple(s):
    v = ast.literal_eval(s)
    if type(v) is not tuple:
        raise argparse.ArgumentTypeError(f"Argument {s} is not a tuple")
    return v


def get_arg_parser(cfg):
    """
    Constructs argument parser based on the fields contained in the flattened
    config with values of form (type, value). The values are set as default fields
    for the parser.
    """
    parser = argparse.ArgumentParser(
        description="Auto-initialized argument parser",
        argument_default=argparse.SUPPRESS,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    for arg, (tp, val) in cfg.items():
        kwargs = {"dest": arg, "help": arg}
        if val is not dataclasses.MISSING:
            kwargs["default"] = val
        if tp is bool:
            kwargs["type"] = str2bool
        elif tp is tuple:
            kwargs["type"] = arg_as_tuple
        else:
            kwargs["type"] = tp
        parser.add_argument(f"--{arg}", **kwargs)

    return parser


def parse_args(cfg, *, cfg_class=None, args=None):
    """
    Main function to parse command line arguments. Returns updated config.

    If `cfg_class` is given, we will convert the parsed config to `cfg_class` after
    parsing is finished.

    If `args=None`, we use the passed command line arguments (sys.argv[1:]). If we
    don't want to parse them, we should set `args=[]`. In this case we will read only
    the config file (if present).
    """
    if args is None:
        args = sys.argv[1:]

    # First we convert all dataclasses, etc. to nested dictionaries
    if cfg_class is None:
        cfg_class = type(cfg) if dataclasses.is_dataclass(cfg) else None
    cfg = to_dict_format(cfg)
    cfg = check_format(cfg)

    # Read config file and apply settings.
    cfg = add_cfg_file_arg(cfg, cfg_class)
    if "cfg_file" in cfg:
        cfg = apply_cfg_file(cfg, args)

    nb_unparsed = len(args)
    unparsed = None
    continue_parsing = nb_unparsed > 0
    while continue_parsing:
        # We do the check at the top, so we do one extra round of parsing to add
        # default args to classes that were potentially specified during the last
        # round of parsing.
        continue_parsing = unparsed is None or len(unparsed) > 0

        # We have to convert to argument format in each iteration, because after calling
        # `parser.parse()` the result is a regular config. We lose type information.
        cfg = to_arg_format(cfg)
        # After having parsed some arguments, we may have gained knowledge about
        # which configurations classes are used. Add those parameters (and defaults)
        # to the config now.
        cfg = add_default_args(cfg)
        # Also add knowledge about arguments from the outer config class
        cfg = add_cls_default_args(cfg, cfg_class)
        # Flatten everything in preparation to parsing.
        cfg = deep_to_flat(cfg)

        # Now we can construct a parser and do some parsing
        parser = get_arg_parser(cfg)
        parsed_cfg, unparsed = parser.parse_known_args(args)
        # After this line `parsed_cfg` will be a flat dictionary.
        parsed_cfg = vars(parsed_cfg)

        # All named parameters in argparse are optional. We don't want that, so we
        # check whether all have been supplied.
        for key in cfg.keys():
            if key not in parsed_cfg:
                raise ValueError(f"Argument {key} was not supplied.")
        cfg = parsed_cfg

        if continue_parsing and len(unparsed) >= nb_unparsed:
            raise ValueError(
                "During the last parsing we have not reduced the number of unparsed "
                "arguments. This suggests that either the user has supplied unknown "
                "arguments or that a '_class' argument is missing.\n"
                f"The list of unparsed arguments is {unparsed}."
            )
        nb_unparsed = len(unparsed)

        # We convert back to nested dictionaries, because adding default arguments
        # wants to work with nested dictionaries.
        cfg = flat_to_deep(cfg)
        cfg = check_format(cfg)

    cfg = to_cls_format(cfg)
    # If the original config was a dataclass, we need to restore the original type
    if cfg_class:
        cfg = cfg_class(**cfg)
    return cfg


def pprint(cfg, indent=2):
    """Nicely prints nested config."""
    cfg = to_dict_format(cfg)
    for key, val in cfg.items():
        if isinstance(val, dict):
            logging.info(" " * indent + str(key) + ":")
            pprint(val, indent + 2)
        else:
            logging.info(" " * indent + str(key) + "=" + str(val))



import numpy as np
import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2_as_graph,
)


def get_flops(model: tf.keras.Model, input_shape: tuple) -> int:
    """
    Calculate FLOPS for a `tf.keras.Model`. Ignore operations used in only training
    mode such as Initialization. Use `tf.profiler` of tensorflow v1 api.

    `input_shape` should be the full input, including batch size.
    """
    if not isinstance(model, tf.keras.Model):
        raise KeyError("`model` argument must be `tf.keras.Model` instance.")

    try:
        # Convert `tf.keras model` into frozen graph
        inputs = tf.TensorSpec(list(input_shape), "float32")
        real_model = tf.function(model).get_concrete_function(inputs)
        frozen_func, _ = convert_variables_to_constants_v2_as_graph(real_model)

        run_meta = tf.compat.v1.RunMetadata()
        opts = tf.compat.v1.profiler.ProfileOptionBuilder.float_operation()
        flops = tf.compat.v1.profiler.profile(
            graph=frozen_func.graph, run_meta=run_meta, cmd="scope", options=opts
        )
        # For some reason `total_float_ops` returns twice the number of FLOPS when
        # compared with results in papers. This difference is consistent. Hence, we
        # divide by 2 to adjust for it, even though we don't understand the reasons...
        flops = flops.total_float_ops // 2
    except ValueError:
        # For big models we can run out of memory...
        flops = 0

    return flops


def get_parameters(model: tf.keras.Model) -> int:
    trainable_params = np.sum([np.prod(v.get_shape()) for v in model.trainable_weights])
    non_trainable_params = np.sum(
        [np.prod(v.get_shape()) for v in model.non_trainable_weights]
    )
    params = trainable_params + non_trainable_params
    return params



import os
from typing import List, Optional

ENV_TFIMM_HOME = "TFIMM_HOME"
ENV_XDG_CACHE_HOME = "XDG_CACHE_HOME"
DEFAULT_CACHE_DIR = "~/.cache"
_tfimm_dir = None
_model_cache = {}


def get_dir() -> str:
    """
    Get the ``tfimm`` cache directory used for storing downloaded models & weights.

    If ``tfimm.set_dir`` has not been called, default path is ``$TFIMM_HOME``, where
    environment variable ``$TFIMM_HOME`` defaults to ``$XDG_CACHE_HOME/tfimm``.
    ``$XDG_CACHE_HOME`` follows the X Design Group specifications of the Linux
    filesystem layout, with a default value ``~/.cache`` if the environment variable is
    not set.
    """
    if _tfimm_dir is not None:
        return _tfimm_dir

    cache_home = os.getenv(ENV_XDG_CACHE_HOME, DEFAULT_CACHE_DIR)
    tfimm_home = os.path.join(cache_home, "tfimm")
    tfimm_home = os.getenv(ENV_TFIMM_HOME, tfimm_home)
    tfimm_home = os.path.expanduser(tfimm_home)
    return tfimm_home


def set_dir(d: str):
    """
    Optionally set the ``tfimm`` models directory used to save downloaded models &
    weights.

    Args:
        d: Path to a local folder to save downloaded models & weights.
    """
    global _tfimm_dir
    _tfimm_dir = d


def set_model_cache(model_name: str, model_path: str):
    """
    Sets the cache location for ``model_name`` to ``model_path``.

    Args:
        model_name: Model name. Same name that is passed to
            :py:func:`~tfimm.create_model`.
        model_path: Path, where the model is stored.
    """
    _model_cache[model_name] = model_path


def clear_model_cache(model_name: str):
    """
    Removes the cache entry for ``model_name``.

    Args:
        model_name: Model name. Same name that is passed to
            :py:func:`~tfimm.create_model`.
    """
    if model_name in _model_cache:
        del _model_cache[model_name]


def list_cached_models() -> List[str]:
    """
    Lists all models in model-specific cache.

    Returns:
        List of model names.
    """
    return sorted(_model_cache.keys())


def cached_model_path(model_name: str) -> Optional[str]:
    """
    Checks if the weights for model ``model_name`` are cached. If so, we return the
    path to the weights, otherwise return ``None``. We check the model-speficif cache
    first before looking in the general cache directory.

    Args:
        model_name: Model to be queried.
    """
    if model_name in _model_cache:
        return _model_cache[model_name]
    else:
        # Now look in the general cache
        model_path = os.path.join(get_dir(), model_name)
        if os.path.exists(model_path):
            return model_path
        else:
            return None



"""
Model Registry

Based on timm/models/registry.py  by Ross Wightman

Copyright 2020 Ross Wightman
Copyright 2021 Martins Bruveris
"""

import fnmatch
import re
import sys
from collections import defaultdict
from copy import deepcopy
from typing import List, Union

__all__ = [
    "list_models",
    "is_model",
    "is_model_in_modules",
    "list_modules",
    "model_class",
    "model_config",
    "register_model",
]

_model_class = {}
_model_config = {}
# Dict of sets to check membership of model in module
_module_to_models = defaultdict(set)
_model_has_pretrained = set()  # Model names that have pretrained weight url present


def register_model(fn):
    # Get model class and model config
    cls, cfg = fn()
    model_name = cfg.name
    if fn.__name__ != model_name:
        raise ValueError(f"Model name({model_name}) != function name ({fn.__name__}).")

    # Lookup module, where model is defined
    mod = sys.modules[fn.__module__]
    module_name_split = fn.__module__.split(".")
    module_name = module_name_split[-1] if len(module_name_split) else ""

    # Add model function to __all__ in that module
    if hasattr(mod, "__all__"):
        mod.__all__.append(model_name)
    else:
        mod.__all__ = [model_name]

    # Add entries to registry dict/sets
    _model_class[model_name] = cls
    _model_config[model_name] = deepcopy(cfg)
    _module_to_models[module_name].add(model_name)
    if cfg.url:  # If URL is non-null, we assume it points to pretrained weights
        _model_has_pretrained.add(model_name)

    return fn


def _natural_key(string_):
    return [int(s) if s.isdigit() else s for s in re.split(r"(\d+)", string_.lower())]


def list_models(
    name_filter: Union[str, List[str]] = "",
    module: str = "",
    pretrained: Union[bool, str] = False,
    exclude_filters: Union[str, List[str]] = "",
):
    """Returns list of available model names, sorted alphabetically.

    Args:
        name_filter: Wildcard filter string that works with fnmatch
        module: Limit model selection to a specific sub-module (ie "resnet")
        pretrained: If True only include models with pretrained weights. If "timm",
            only include models with pretrained weights in timm library
        exclude_filters (str or list[str]) - Wildcard filters to exclude models after
            including them with filter

    Example:
        model_list("gluon_resnet*") -- returns all models starting with "gluon_resnet"
        model_list("*resnext*", "resnet") -- returns all models with "resnext" in
            "resnet" module
    """
    if module:
        all_models = list(_module_to_models[module])
    else:
        all_models = _model_class.keys()

    if name_filter:
        if not isinstance(name_filter, (tuple, list)):
            name_filter = [name_filter]
        models = set()
        for f in name_filter:
            include_models = fnmatch.filter(all_models, f)  # include these models
            if len(include_models):
                models = models.union(include_models)
    else:
        models = set(all_models)

    if exclude_filters:
        if not isinstance(exclude_filters, (tuple, list)):
            exclude_filters = [exclude_filters]
        for xf in exclude_filters:
            exclude_models = fnmatch.filter(models, xf)  # exclude these models
            if len(exclude_models):
                models = models.difference(exclude_models)

    if pretrained is True:
        models = _model_has_pretrained.intersection(models)
    elif pretrained == "timm":
        models = models.intersection(_timm_pretrained_models())

    return list(sorted(models, key=_natural_key))


def is_model(model_name):
    """Check if a model name exists"""
    return model_name in _model_class


def model_class(model_name):
    """Fetch a model entrypoint for specified model name"""
    return _model_class[model_name]


def model_config(model_name):
    """Fetch a model config for specified model name"""
    return _model_config[model_name]


def list_modules():
    """Return list of module names that contain models / model entrypoints"""
    modules = _module_to_models.keys()
    return list(sorted(modules))


def is_model_in_modules(model_name, module_names):
    """Check if a model exists within a subset of modules
    Args:
        model_name (str) - name of model to check
        module_names (tuple, list, set) - names of modules to search in
    """
    assert isinstance(module_names, (tuple, list, set))
    return any(model_name in _module_to_models[n] for n in module_names)


def is_model_pretrained(model_name):
    return model_name in _model_has_pretrained


def _timm_pretrained_models():
    """Returns list of models with pretrained weights in timm library."""
    import timm

    models = timm.list_models(pretrained=True)
    return set(models)


def _to_timm_module_name(module):
    """
    Some modules are called differently in tfimm and timm. This function converts the
    tfimm name to the timm name.
    """
    if module == "vit":
        module = "vision_transformer"
    elif module == "swin":
        module = "swin_transformer"
    return module


# TODO: Make it work with efficientnet, i.e., diverging names between TFIMM and TIMM
def _compare_available_models_with_timm(
    name_filter: Union[str, List[str]] = "",
    module: str = "",
    exclude_filters: Union[str, List[str]] = "",
):
    """Helper function to list which models have not yet been converted from timm."""
    import timm

    tf_models = list_models(
        name_filter=name_filter,
        module=module,
        pretrained="timm",
        exclude_filters=exclude_filters,
    )
    pt_models = timm.list_models(
        filter=name_filter,
        module=_to_timm_module_name(module),
        pretrained=True,
        exclude_filters=exclude_filters,
    )

    pt_only = sorted(list(set(pt_models) - set(tf_models)))
    print(f"timm models available in tfimm: {len(tf_models)}/{len(pt_models)}.")
    print(f"timm models not available: {len(pt_only)}.")
    print(f"The following models are not available: {', '.join(pt_only)}")
